@@ -1,6 +1,7 @@
 """IPython terminal interface using prompt_toolkit in place of readline"""
 from __future__ import print_function
 
+import asyncio
 import base64
 import errno
 from getpass import getpass
@@ -13,34 +14,61 @@ import sys
 import time
 from warnings import warn
 
+from typing import Dict as DictType, Any as AnyType
+
 from zmq import ZMQError
 from IPython.core import page
 from ipython_genutils.tempdir import NamedFileInTemporaryDirectory
-from traitlets import (Bool, Integer, Float, Unicode, List, Dict, Enum,
-                       Instance, Any)
+from traitlets import (
+    Bool,
+    Integer,
+    Float,
+    Unicode,
+    List,
+    Dict,
+    Enum,
+    Instance,
+    Any,
+)
 from traitlets.config import SingletonConfigurable
 
 from .completer import ZMQCompleter
 from .zmqhistory import ZMQHistoryManager
 from . import __version__
 
+# Discriminate version3 for asyncio
+from prompt_toolkit import __version__ as ptk_version
+PTK3 = ptk_version.startswith('3.')
+
+if not PTK3:
+    # use_ayncio_event_loop obsolete in PKT3
+    from prompt_toolkit.eventloop.defaults import use_asyncio_event_loop
+
 from prompt_toolkit.completion import Completer, Completion
 from prompt_toolkit.document import Document
 from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
-from prompt_toolkit.filters import (Condition, has_focus, has_selection,
-                                    vi_insert_mode, emacs_insert_mode, is_done)
+from prompt_toolkit.filters import (
+    Condition,
+    has_focus,
+    has_selection,
+    vi_insert_mode,
+    emacs_insert_mode,
+    is_done,
+)
 from prompt_toolkit.history import InMemoryHistory
 from prompt_toolkit.patch_stdout import patch_stdout
 from prompt_toolkit.shortcuts.prompt import PromptSession
-from prompt_toolkit.shortcuts import print_formatted_text
+from prompt_toolkit.shortcuts import print_formatted_text, CompleteStyle
 from prompt_toolkit.key_binding import KeyBindings
 from prompt_toolkit.lexers import PygmentsLexer
-from prompt_toolkit.layout.processors import (ConditionalProcessor,
-                                              HighlightMatchingBracketProcessor)
+from prompt_toolkit.layout.processors import (
+    ConditionalProcessor,
+    HighlightMatchingBracketProcessor,
+)
 from prompt_toolkit.styles import merge_styles
 from prompt_toolkit.styles.pygments import (style_from_pygments_cls,
                                             style_from_pygments_dict)
-from prompt_toolkit.formatted_text import PygmentsTokens
+from prompt_toolkit.formatted_text import PygmentsTokens, FormattedText
 from prompt_toolkit.output import ColorDepth
 from prompt_toolkit.utils import suspend_to_background_supported
 
@@ -83,6 +111,15 @@ def ask_yes_no(prompt, default=None, interrupt=None):
     return answers[ans]
 
 
+async def async_input(prompt, loop=None):
+    """Simple async version of input using a the default executor"""
+    if loop is None:
+        loop = asyncio.get_event_loop()
+
+    raw = await loop.run_in_executor(None, input, prompt)
+    return raw
+
+
 def get_pygments_lexer(name):
     name = name.lower()
     if name == 'ipython2':
@@ -113,9 +150,26 @@ class JupyterPTCompleter(Completer):
             code=document.text,
             cursor_pos=document.cursor_position
         )
-        start_pos = content['cursor_start'] - document.cursor_position
-        for m in content['matches']:
-            yield Completion(m, start_position=start_pos)
+        meta = content["metadata"]
+
+        if "_jupyter_types_experimental" in meta:
+            try:
+                new_meta = {}
+                for c, m in zip(
+                    content["matches"], meta["_jupyter_types_experimental"]
+                ):
+                    new_meta[c] = m["type"]
+                meta = new_meta
+            except:
+                pass
+
+        start_pos = content["cursor_start"] - document.cursor_position
+        for m in content["matches"]:
+            yield Completion(
+                m,
+                start_position=start_pos,
+                display_meta=meta.get(m, "?"),
+            )
 
 
 class ZMQTerminalInteractiveShell(SingletonConfigurable):
@@ -206,8 +260,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         """
     )
 
-    callable_image_handler = Any(config=True, help=
-        """
+    callable_image_handler = Any(
+        config=True,
+        help="""
         Callable object called via 'callable' image handler with one
         argument, `data`, which is `msg["content"]["data"]` where
         `msg` is the message from iopub channel.  For example, you can
@@ -248,15 +303,29 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         help="""Set to display confirmation dialog on exit.
         You can always use 'exit' or 'quit', to force a
         direct exit without any confirmation.
-        """
+        """,
     )
 
-    highlight_matching_brackets = Bool(True,
-        help="Highlight matching brackets.",
+    display_completions = Enum(
+        ("column", "multicolumn", "readlinelike"),
+        help=(
+            "Options for displaying tab completions, 'column', 'multicolumn', and "
+            "'readlinelike'. These options are for `prompt_toolkit`, see "
+            "`prompt_toolkit` documentation for more information."
+        ),
+        default_value="multicolumn",
+    ).tag(config=True)
+    
+    prompt_includes_vi_mode = Bool(True,
+        help="Display the current vi mode (when using vi editing mode)."
     ).tag(config=True)
 
-    manager = Instance('jupyter_client.KernelManager', allow_none=True)
-    client = Instance('jupyter_client.KernelClient', allow_none=True)
+    highlight_matching_brackets = Bool(True, help="Highlight matching brackets.",).tag(
+        config=True
+    )
+
+    manager = Instance("jupyter_client.KernelManager", allow_none=True)
+    client = Instance("jupyter_client.KernelClient", allow_none=True)
 
     def _client_changed(self, name, old, new):
         self.session_id = new.session.session
@@ -299,16 +368,25 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         self.history_manager = ZMQHistoryManager(client=self.client)
         self.configurables.append(self.history_manager)
 
-    def get_prompt_tokens(self):
+    def vi_mode(self):
+        if (getattr(self, 'editing_mode', None) == 'vi'
+                and self.prompt_includes_vi_mode):
+            return '['+str(self.pt_cli.app.vi_state.input_mode)[3:6]+'] '
+        return ''
+    
+    def get_prompt_tokens(self, ec=None):
+        if ec is None:
+            ec = self.execution_count
         return [
+            (Token.Prompt, self.vi_mode()),
             (Token.Prompt, 'In ['),
-            (Token.PromptNum, str(self.execution_count)),
+            (Token.PromptNum, str(ec)),
             (Token.Prompt, ']: '),
         ]
 
     def get_continuation_tokens(self, width):
         return [
-            (Token.Prompt, (' ' * (width - 2)) + ': '),
+            (Token.Prompt, (" " * (width - 5)) + "...: "),
         ]
 
     def get_out_prompt_tokens(self):
@@ -323,7 +401,26 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         print_formatted_text(PygmentsTokens(tokens), end='',
                              style = self.pt_cli.app.style)
 
-    kernel_info = {}
+    def get_remote_prompt_tokens(self):
+        return [
+            (Token.RemotePrompt, self.other_output_prefix),
+        ]
+
+    def print_remote_prompt(self, ec=None):
+        tokens = self.get_remote_prompt_tokens() + self.get_prompt_tokens(ec=ec)
+        print_formatted_text(
+            PygmentsTokens(tokens), end="", style=self.pt_cli.app.style
+        )
+
+    @property
+    def pt_complete_style(self):
+        return {
+            "multicolumn": CompleteStyle.MULTI_COLUMN,
+            "column": CompleteStyle.COLUMN,
+            "readlinelike": CompleteStyle.READLINE_LIKE,
+        }[self.display_completions]
+
+    kernel_info: DictType[str, AnyType] = {}
 
     def init_kernel_info(self):
         """Wait for a kernel to be ready, and store kernel info"""
@@ -334,9 +431,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         while True:
             try:
                 reply = self.client.get_shell_msg(timeout=1)
-            except Empty:
+            except Empty as e:
                 if (time.time() - tic) > timeout:
-                    raise RuntimeError("Kernel didn't respond to kernel_info_request")
+                    raise RuntimeError("Kernel didn't respond to kernel_info_request") from e
             else:
                 if reply['parent_header'].get('msg_id') == msg_id:
                     self.kernel_info = reply['content']
@@ -350,8 +447,10 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         if self.simple_prompt or ('JUPYTER_CONSOLE_TEST' in os.environ):
             # Simple restricted interface for tests so we can find prompts with
             # pexpect. Multi-line input not supported.
-            def prompt():
-                return input('In [%d]: ' % self.execution_count)
+            async def prompt():
+                prompt = 'In [%d]: ' % self.execution_count
+                raw = await async_input(prompt)
+                return raw
             self.prompt_for_code = prompt
             self.print_out_prompt = \
                 lambda: print('Out[%d]: ' % self.execution_count, end='')
@@ -395,6 +494,10 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         def _(event):
             event.cli.suspend_to_background()
 
+        @kb.add("c-o", filter=(has_focus(DEFAULT_BUFFER) & emacs_insert_mode))
+        def _(event):
+            event.current_buffer.insert_text("\n")
+
         # Pre-populate history from IPython's history database
         history = InMemoryHistory()
         last_cell = u""
@@ -410,6 +513,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             Token.PromptNum: '#00ff00 bold',
             Token.OutPrompt: '#ff2200',
             Token.OutPromptNum: '#ff0000 bold',
+            Token.RemotePrompt: '#999900',
         }
         if self.highlighting_style:
             style_cls = get_style_by_name(self.highlighting_style)
@@ -444,14 +548,21 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             Condition(lambda: self.highlight_matching_brackets))
         ]
 
+        # Tell prompt_toolkit to use the asyncio event loop.
+        # Obsolete in prompt_toolkit.v3
+        if not PTK3:
+            use_asyncio_event_loop()
+
         self.pt_cli = PromptSession(
             message=(lambda: PygmentsTokens(self.get_prompt_tokens())),
             multiline=True,
+            complete_style=self.pt_complete_style,
             editing_mode=editing_mode,
             lexer=PygmentsLexer(get_pygments_lexer(lexer)),
             prompt_continuation=(
-                lambda width, lineno, is_soft_wrap:
-                PygmentsTokens(self.get_continuation_tokens(width))
+                lambda width, lineno, is_soft_wrap: PygmentsTokens(
+                    self.get_continuation_tokens(width)
+                )
             ),
             key_bindings=kb,
             history=history,
@@ -460,21 +571,20 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             style=style,
             input_processors=input_processors,
             color_depth=(ColorDepth.TRUE_COLOR if self.true_color else None),
-
         )
 
-    def prompt_for_code(self):
+    async def prompt_for_code(self):
         if self.next_input:
             default = self.next_input
             self.next_input = None
         else:
             default = ''
 
-        with patch_stdout(raw=True):
-            text = self.pt_cli.prompt(
-                default=default,
-#                pre_run=self.pre_prompt,# reset_current_buffer=True,
-            )
+        if PTK3:
+            text = await self.pt_cli.prompt_async(default=default)
+        else:
+            text = await self.pt_cli.prompt(default=default, async_=True)
+
         return text
 
     def init_io(self):
@@ -523,12 +633,12 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                 set_doc()
             self.next_input = None
 
-    def interact(self, display_banner=None):
+    async def interact(self, loop=None, display_banner=None):
         while self.keep_running:
             print('\n', end='')
 
             try:
-                code = self.prompt_for_code()
+                code = await self.prompt_for_code()
             except EOFError:
                 if (not self.confirm_exit) or \
                         ask_yes_no('Do you really want to exit ([y]/n)?', 'y', 'n'):
@@ -540,11 +650,29 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
 
     def mainloop(self):
         self.keepkernel = not self.own_kernel
+        loop = asyncio.get_event_loop()
         # An extra layer of protection in case someone mashing Ctrl-C breaks
         # out of our internal code.
         while True:
             try:
-                self.interact()
+                tasks = [self.interact(loop=loop)]
+
+                if self.include_other_output:
+                    # only poll the iopub channel asynchronously if we
+                    # wish to include external content
+                    tasks.append(self.handle_external_iopub(loop=loop))
+
+                main_task = asyncio.wait(tasks, loop=loop, return_when=asyncio.FIRST_COMPLETED)
+                _, pending = loop.run_until_complete(main_task)
+
+                for task in pending:
+                    task.cancel()
+                try:
+                    loop.run_until_complete(asyncio.gather(*pending))
+                except asyncio.CancelledError:
+                    pass
+                loop.stop()
+                loop.close()
                 break
             except KeyboardInterrupt:
                 print("\nKeyboardInterrupt escaped interact()\n")
@@ -680,11 +808,9 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
     include_other_output = Bool(False, config=True,
         help="""Whether to include output from clients
         other than this one sharing the same kernel.
-
-        Outputs are not displayed until enter is pressed.
         """
     )
-    other_output_prefix = Unicode("[remote] ", config=True,
+    other_output_prefix = Unicode("Remote ", config=True,
         help="""Prefix to add to outputs coming from clients other than this one.
 
         Only relevant if include_other_output is True.
@@ -707,6 +833,15 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         else:
             return from_here
 
+    async def handle_external_iopub(self, loop=None):
+        while self.keep_running:
+            # we need to check for keep_running from time to time as
+            # we are blocking in an executor block which cannot be cancelled.
+            poll_result = await loop.run_in_executor(
+                None, self.client.iopub_channel.socket.poll, 500)
+            if(poll_result):
+                self.handle_iopub()
+
     def handle_iopub(self, msg_id=''):
         """Process messages on the IOPub channel
 
@@ -727,6 +862,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             if self.include_output(sub_msg):
                 if msg_type == 'status':
                     self._execution_state = sub_msg["content"]["execution_state"]
+
                 elif msg_type == 'stream':
                     if sub_msg["content"]["name"] == "stdout":
                         if self._pending_clearoutput:
@@ -765,6 +901,12 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                         print()
                     print(text_repr)
 
+                    # Remote: add new prompt
+                    if not self.from_here(sub_msg):
+                        sys.stdout.write('\n')
+                        sys.stdout.flush()
+                        self.print_remote_prompt()
+
                 elif msg_type == 'display_data':
                     data = sub_msg["content"]["data"]
                     handled = self.handle_rich_data(data)
@@ -775,11 +917,19 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
                         if 'text/plain' in data:
                             print(data['text/plain'])
 
+                # If execute input: print it
                 elif msg_type == 'execute_input':
                     content = sub_msg['content']
-                    if not self.from_here(sub_msg):
-                        sys.stdout.write(self.other_output_prefix)
-                    sys.stdout.write('In [{}]: '.format(content['execution_count']))
+                    ec = content.get('execution_count', self.execution_count - 1)
+
+                    # New line
+                    sys.stdout.write('\n')
+                    sys.stdout.flush()
+
+                    # With `Remote In [3]: `
+                    self.print_remote_prompt(ec=ec)
+
+                    # And the code
                     sys.stdout.write(content['code'] + '\n')
 
                 elif msg_type == 'clear_output':
@@ -818,12 +968,12 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
             from PIL import Image, ImageShow
         except ImportError:
             return False
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         img = Image.open(BytesIO(raw))
         return ImageShow.show(img)
 
     def handle_image_stream(self, data, mime):
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         imageformat = self._imagemime[mime]
         fmt = dict(format=imageformat)
         args = [s.format(**fmt) for s in self.stream_image_handler]
@@ -835,7 +985,7 @@ class ZMQTerminalInteractiveShell(SingletonConfigurable):
         return (proc.returncode == 0)
 
     def handle_image_tempfile(self, data, mime):
-        raw = base64.decodestring(data[mime].encode('ascii'))
+        raw = base64.decodebytes(data[mime].encode('ascii'))
         imageformat = self._imagemime[mime]
         filename = 'tmp.{0}'.format(imageformat)
         with NamedFileInTemporaryDirectory(filename) as f, \
